/* mkchain.c */
/* Copyright (C) 2017 Norbert Doerry
   May 2017


   MKCHAIN.EXE xmfile.xml 

   xmfile.xml = input file - required


   creates a comma separated value (CSV) file where the 
   rows are nominally years (or time increments)
   the columns are each a markov chain of states (represented as integers)
   The input xml file defines the probability for the initial state and the state transition matrix

++++++++++

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "xml.h"
#include "mkchain.h"

void init_mkchain(MKCHAIN *);
void read_xml_file(MKCHAIN *);
void printheader(FILE *,MKCHAIN *);
void printmkchain(FILE *,MKCHAIN *);
void error_check_mkchain(MKCHAIN *);
void create_mkchain(int,MKCHAIN *);
int calc_state(double *, int );
int calc_steady_state(double **,double *, int , int, double, double *,int);
void multiply_array_by_array(double **, double **, double **,int,int,int);


int main(int argc, char **argv)
{
 MKCHAIN mk;
 int i;

 init_mkchain(&mk);

 if (argc < 2)
  {
   fprintf(stderr," *** ERROR: input xml filename not specified\n");
   exit(EXIT_FAILURE);
  }

 mk.xmlfilename = argv[1];
 mk.in = fopen(mk.xmlfilename,"r");
 if (mk.in == (FILE *) NULL)
  {
   fprintf(stderr," *** ERROR: Unable to open file (%s)\n",mk.xmlfilename);
   exit(EXIT_FAILURE);
  }


 read_xml_file(&mk);

 if (mk.debug){printf("\n*"); fflush(stdout);}

 error_check_mkchain(&mk);
 
 for (i = 1 ; i <= mk.nbr_chains ; i++)
  create_mkchain(i,&mk);

 if (mk.debug){printf("*");fflush(stdout);}

 if (mk.calc_flag)
  {
   mk.actual_iteration = calc_steady_state(mk.transition_probability,mk.steady_state_probability, 
	mk.nbr_states, mk.max_iteration, mk.max_error, &(mk.actual_error),mk.debug);
  }
 if (mk.debug){printf("*");fflush(stdout);}

 if (mk.print_header != 0)
  printheader(mk.out,&mk);

 if (mk.debug){printf("*");fflush(stdout);}

 printmkchain(mk.out,&mk);

 if (mk.debug){printf("*");fflush(stdout);}

 if (mk.out != stdout) fclose(mk.out);

 return EXIT_SUCCESS;
}


void init_mkchain(MKCHAIN *mk)
{
 if (mk == (MKCHAIN *) NULL)
  exit(EXIT_FAILURE);

 mk->nbr_states = 0;
 mk->init_year = 0;
 mk->nbr_years = 0;
 mk->nbr_chains = 0;
 mk->outfilename = (char *) NULL;
 mk->out = (FILE *) NULL;
 mk->xmlfilename = (char *) NULL;
 mk->in = (FILE *) NULL;
 mk->state_name = (char **) NULL;
 mk->val = (double *) NULL;
 mk->initial_probability = (double *) NULL;
 mk->transition_probability = (double **) NULL;

 mk->print_header = 0;
 mk->print_flag = PRINT_VAL;

 mk->steady_state_probability = (double *) NULL;
 mk->calc_flag = 1;
 mk->max_iteration = 0;
 mk->max_error = 0.0;

 mk->debug = 0;

}

void read_xml_file(MKCHAIN *mk)
{
 XMLENTITY *xml, *xp,*xp1,*xp2,*xp3;
 int errflag;
 int i;
 int ser, ser_to;
 char *name;
 double initial_probability, probability;
 double val;


 if (mk->in == (FILE *) NULL)
  {
   fprintf(stderr," *** ERROR: NULL pointer for FILE mk->in in read_xml_file()\n");
   exit(EXIT_FAILURE);
  }


 xml = readfile_XMLENTITY(mk->in,&errflag);

 if (mk->in != stdin) fclose(mk->in);

 if (errflag != ERROR_NONE || xml == (XMLENTITY *) NULL)
  {
    fprintf(stderr," *** ERROR: (%d) Error encountered in reading XML file\n",errflag);
    exit(EXIT_FAILURE);
  }


 name2lower_XMLENTITY(xml,1);


 mk->xml = xml;
 
 if (strcmp(xml->name,"mkchain") != 0)
  {
   fprintf(stderr," *** ERROR:  (%s) Not a mkchain XML file\n", xml->name);
   exit(EXIT_FAILURE);
  }

 for (xp = xml->child;  xp != (XMLENTITY *) NULL ; xp = xp->next)
  {
    
    if (xp->type != ENTITY_DATA)
     continue;

    if (strncmp(xp->name,"nbr_states",8) == 0)
     {
       mk->nbr_states = atoi(xp->s);
     }
    else if (strncmp(xp->name,"init_year",8) == 0)
     {
      mk->init_year = atoi(xp->s);
     }
    else if (strncmp(xp->name,"nbr_years",8) == 0)
     {
      mk->nbr_years = atoi(xp->s);
     }
    else if (strncmp(xp->name,"nbr_chains",8) == 0)
     {
      mk->nbr_chains = atoi(xp->s);   
     }
    else if (strncmp(xp->name,"print_flag",8) == 0)
     {
      if (strcmp(xp->s,"VAL") == 0 || strcmp(xp->s,"val") == 0)
       mk->print_flag = PRINT_VAL;
      else if (strcmp(xp->s,"SER") == 0 || strcmp(xp->s,"ser") == 0)
       mk->print_flag = PRINT_SER;
      else if (strcmp(xp->s,"NAME") == 0 || strcmp(xp->s,"name") == 0)
       mk->print_flag = PRINT_NAME;
     }
    else if (strncmp(xp->name,"print_header",8) == 0)
     {
       if (strcmp(xp->s,"YES") == 0 || strcmp(xp->s,"yes") == 0 || strcmp(xp->s,"1") == 0)
        mk->print_header = 1;
       else if (strcmp(xp->s,"NO") == 0 || strcmp(xp->s,"no") == 0 || strcmp(xp->s,"0") == 0)
        mk->print_header = 0;

     }
    else if (strcmp(xp->name,"debug") == 0)
     {
       mk->debug = atoi(xp->s);
     }
    else if (strcmp(xp->name,"filename") == 0)
     {
        mk->outfilename = xp->s;
     }

    else if (strncmp(xp->name,"steadystate",8) == 0)
     {
        for (xp1 = xp->child ; xp1 != (XMLENTITY *) NULL ; xp1 = xp1->next)
         {
          if (strcmp(xp1->name,"calc") == 0)
            {
             if (strcmp(xp1->s,"YES") == 0 || strcmp(xp1->s,"yes") == 0 || strcmp(xp1->s,"1") == 0)
              mk->calc_flag = 1;
             else if (strcmp(xp1->s,"NO") == 0 || strcmp(xp1->s,"no") == 0 || strcmp(xp1->s,"0") == 0)
              mk->calc_flag = 0;
            }
          else if (strncmp(xp1->name,"max_iteration",8) == 0)
           {
            mk->max_iteration = atoi(xp1->s);
           }
          else if (strncmp(xp1->name,"max_error",8) == 0)
           {
            mk->max_error = atof(xp1->s);
           }
          else
           {
            fprintf(stderr," *** WARNING: Unknown tag (%s) ignored\n",xp1->name);
           }
         }
     }

    /* skip "state" and "transition" for now */
 
    else if (strcmp(xp->name,"state") == 0)
      continue;
    else if (strcmp(xp->name,"transition") == 0)
      continue;
    else
     {
       fprintf(stderr," *** WARNING: Unknown tag (%s) ignored\n",xp->name);
     }

  }

 /* error checking */

 if (mk->nbr_states < 2)
  {
    fprintf(stderr, " *** ERROR: nbr_states (%d) must be greater than 1\n",mk->nbr_states);
    exit(EXIT_FAILURE);
  }

 if (mk->nbr_years < 2)
  mk->nbr_years = DEFAULT_NBR_YEARS;

 if (mk->nbr_chains < 1)
  mk->nbr_chains = DEFAULT_NBR_CHAINS;


 if (mk->max_iteration < 1)
  mk->max_iteration = DEFAULT_MAX_ITERATION;

 if (mk->max_error <= 0.0)
  mk->max_error = DEFAULT_MAX_ERROR;

 if (mk->outfilename == (char *) NULL)
  {
    mk->out == stdout;
  }
 else
  {
    mk->out = fopen(mk->outfilename,"w");
    if (mk->out == (FILE *) NULL)
     {
       fprintf(stderr, "*** ERROR: Unable to open file (%s) for writing, using stdout\n",mk->outfilename);
       mk->out= stdout;
     }

  } 
  
 /* allocate arrays */


 mk->state_name = (char **)calloc((size_t)mk->nbr_states+1,sizeof(char *));
 mk->val = (double *) calloc((size_t)mk->nbr_states+1,sizeof(double));
 mk->initial_probability = (double *) calloc((size_t)mk->nbr_states+1,sizeof(double));
 mk->transition_probability = (double **) calloc((size_t)mk->nbr_states+1,sizeof(double *));
 mk->mkchain = (int **) calloc((size_t)mk->nbr_chains+1,sizeof(int *));
 mk->steady_state_probability = (double *) calloc((size_t)mk->nbr_states+1,sizeof(double));

 if (mk->state_name == (char **) NULL || mk->val == (double *)NULL || mk->initial_probability == (double *) NULL ||
     mk->transition_probability == (double **) NULL || mk->mkchain == (int **) NULL || mk->steady_state_probability == (double *) NULL)
  {
   fprintf(stderr," *** ERROR: out of memory 1 in read_xml_file()\n");
   exit(EXIT_FAILURE);

  }

 for (i = 0 ; i <= mk->nbr_states ; i++)
  {
    mk->transition_probability[i] = (double *) calloc((size_t)mk->nbr_states+1,sizeof(double));
    if (mk->transition_probability[i] == (double *) NULL)
     {
      fprintf(stderr," *** ERROR: out of memory 2 in read_xml_file()\n");
      exit(EXIT_FAILURE);
     }
  }

 for (i = 0 ; i <= mk->nbr_chains ; i++)
  {
    mk->mkchain[i] = (int *) calloc((size_t)mk->nbr_years+1,sizeof(int));
    if (mk->mkchain[i] == (int *) NULL)
     {
      fprintf(stderr," *** ERROR: out of memory 3 in read_xml_file()\n");
      exit(EXIT_FAILURE);
     }
  }
 
 for (xp = xml->child;  xp != (XMLENTITY *) NULL ; xp = xp->next)
  {

    if (strcmp(xp->name,"state") == 0)
      {

        ser = 0;
        name = (char *) NULL;
        initial_probability = 0.0;
        val = 0.0;

        for (xp1 = xp->child ; xp1 != (XMLENTITY *) NULL ; xp1 = xp1->next)
         {
           if (strcmp(xp1->name,"ser") == 0)
            ser = atoi(xp1->s);
           else if (strcmp(xp1->name,"name") == 0)
            name = xp1->s;
           else if (strncmp(xp1->name,"initial_probability",8) == 0)
            initial_probability = atof(xp1->s);
           else if (strncmp(xp1->name,"value",3) == 0)
            val = atof(xp1->s);
           else
            {
             fprintf(stderr," *** WARNING: Unknown tag (%s) ignored\n",xp1->name);
            }
         }

        if (ser < 1 || ser > mk->nbr_states)
         {

           fprintf(stderr," *** ERROR: For STATE (%s) SER (%d) is < 1 or > NBR_STATES (%d)\n",name,ser,mk->nbr_states);
           exit(EXIT_FAILURE);

         }
        if (initial_probability < 0.0 ||initial_probability > 1.0)
         {
           fprintf(stderr," *** ERROR: For STATE (%s) illegal INITIAL_PROBABILITY (%f)\n",name,initial_probability);
           exit(EXIT_FAILURE);
         }

        mk->state_name[ser] = name;
        mk->val[ser] = val;
        mk->initial_probability[ser] = initial_probability;

      }
    else if (strcmp(xp->name,"transition") == 0)
      {
        ser = 0;

        for (xp1 = xp->child ; xp1 != (XMLENTITY *) NULL ; xp1 = xp1->next)
         {
           if (strcmp(xp1->name,"from") == 0)
            ser = atoi(xp1->s);
           else if (strcmp(xp1->name,"to") == 0)
            continue;
           else
            {
             fprintf(stderr," *** WARNING: Unknown tag (%s) ignored\n",xp1->name);
            }
          }

        if (ser < 1 || ser > mk->nbr_states)
         {

           fprintf(stderr," *** ERROR: TRANSITION FROM (%d) is < 1 or > NBR_STATES (%d)\n",ser,mk->nbr_states);
           exit(EXIT_FAILURE);

         }
        for (xp1 = xp->child ; xp1 != (XMLENTITY *) NULL ; xp1 = xp1->next)
         {
           if (strcmp(xp1->name,"to") != 0) continue;

           for (xp2 = xp1->child ; xp2 != (XMLENTITY *) NULL ; xp2 = xp2->next)
            {
              if (strcmp(xp2->name,"state") == 0)
               {
                ser_to = atoi(xp2->s);
               }
              else if (strncmp(xp2->name,"probability",8) == 0)
               {
                probability = atof(xp2->s);
               }
              else 
               {
                fprintf(stderr," *** WARNING: Unknown tag (%s) ignored\n",xp2->name);
               }

             }
            if (ser_to < 1 || ser_to > mk->nbr_states)
             {
              fprintf(stderr," *** ERROR: TRANSITION FROM (%d) TO (%d) is < 1 or > NBR_STATES (%d)\n",
		ser, ser_to,mk->nbr_states);
               exit(EXIT_FAILURE);
              }

            if (probability < 0.0 || probability > 1.0)
             {
               fprintf(stderr," *** ERROR: For TRANSITION FROM (%d) TO (%d)  illegal PROBABILITY (%f)\n",
		ser,ser_to,probability);
               exit(EXIT_FAILURE);
             }
            
            mk->transition_probability[ser][ser_to] = probability;

         }
      }


  }

}

/* normalize the probabilities and check to ensure probabilities are non-zero */

void error_check_mkchain(MKCHAIN *mk)
{

 if(mk == (MKCHAIN *) NULL)
  {
   fprintf(stderr," *** ERROR: NULL pointer for mk in error_check_mkchain()\n");
   exit(EXIT_FAILURE);
  }

 double prob_sum;
 int i,j;

 /* check for missing state names */

 for (i=1 ; i <= mk->nbr_states ; i++)
  {
    if (mk->state_name[i] == (char *) NULL)
     fprintf(stderr," *** WARNING: Missing NAME for STATE (%d)\n",i);
  }

 /* normalize probabilities */

 prob_sum = 0.0;
 for (i=1 ; i <= mk->nbr_states ; i++)
  prob_sum += mk->initial_probability[i];

 if (prob_sum <= 0.0)
  {
   fprintf(stderr," *** ERROR: zero total initial_probability for all states\n");
   exit(EXIT_FAILURE);
  }

 /* normalize probabilities if they don't sum to 1.0 */

 if (prob_sum != 1.0)
  {
   fprintf(stderr," *** WARNING: normalizing initial_probability for all states by %f\n",prob_sum);
   for (i=1 ; i <= mk->nbr_states ; i++)
    mk->initial_probability[i] = mk->initial_probability[i] / prob_sum;

  }

 /* check transition probabilities, each row should sum to a number greater than 0 */

 for (i = 1 ; i <= mk->nbr_states ; i++)
  {
    prob_sum = 0.0;
    
    for (j = 1 ; j <= mk->nbr_states ; j++) 
     prob_sum += mk->transition_probability[i][j];

    if (prob_sum <= 0.0)
     {
      fprintf(stderr," *** ERROR: zero total transition_probability for state (%d)\n",i);
       exit(EXIT_FAILURE);
     }
 
    if (prob_sum != 1.0)
     {
       fprintf(stderr," *** WARNING: normalizing transition_probability for state (%d) by %f\n",i,prob_sum);
       for (j=1 ; j <= mk->nbr_states ; j++)
         mk->transition_probability[i][j] = mk->transition_probability[i][j] / prob_sum;
     }

  }

}


void printheader(FILE *out,MKCHAIN *mk)
{
 time_t tm;
 char *tm_ptr;
 int i,j;

 if (mk == (MKCHAIN *) NULL) return;

 if (out == (FILE *) NULL)
  out = stdout;

 time(&tm);
 tm_ptr = ctime(&tm); 

 fprintf(out,"MKCHAIN VERSION , %s\n", VERSION);
 fprintf(out,"time , %s",tm_ptr);
 

 if (mk->xmlfilename != (char *) NULL)
  fprintf(out,"xmlfilename , \"%s\"\n",mk->xmlfilename);
 else
  fprintf(out,"xmlfilename\n");

 if (mk->outfilename != (char *) NULL)
  fprintf(out,"outfilename , \"%s\"\n",mk->outfilename);
 else
  fprintf(out,"outfilename\n");

 fprintf(out,"nbr_states , %d\n",mk->nbr_states);
 fprintf(out,"init_year  , %d\n",mk->init_year);
 fprintf(out,"nbr_years  , %d\n",mk->nbr_years);
 fprintf(out,"nbr_chains , %d\n",mk->nbr_chains);
 fprintf(out,"state , name , val , initial_probability\n");
 for (i = 1 ; i <= mk->nbr_states ; i++)
  fprintf(out,"%d , \"%s\" , %f , %f\n",i,mk->state_name[i],mk->val[i],mk->initial_probability[i]);
 fprintf(out,"transition_probability\n");
 for (i = 1 ; i <= mk->nbr_states ; i++)
  {
    for (j = 1 ; j < mk->nbr_states ; j++)
     {
       fprintf(out," %f ,",mk->transition_probability[j][i]);
     }
    fprintf(out," %f\n",mk->transition_probability[j][i]);
  }

 if (mk->calc_flag)
 {
   fprintf(out,"max_iteration , %d , actual_iteration, %d\n",mk->max_iteration,mk->actual_iteration);
   fprintf(out,"max_error , %f , actual_error, %f\n",mk->max_error, mk->actual_error);

   fprintf(out,"state ,  steady_state_probability\n");
   for (i = 1 ; i <= mk->nbr_states ; i++)
    fprintf(out,"%d , %f\n",i,mk->steady_state_probability[i]);
  }

}

void printmkchain(FILE *out,MKCHAIN *mk)
{
int i,j;

 if (mk == (MKCHAIN *) NULL) return;

 if (out == (FILE *) NULL)
  out = stdout;

 fprintf(out,"time ");

 for (i = 1 ; i <= mk->nbr_chains ; i++)
  fprintf(out,", %d ",i);

 fprintf(out,"\n");

 fflush(out);

 for (j = 0 ; j <= mk->nbr_years ; j++)
  {
    fprintf(out," %d ",mk->init_year + j);

    for (i = 1 ; i <= mk->nbr_chains ; i++)
     {

      if(mk->print_flag == PRINT_VAL)
       fprintf(out,", %f ",mk->val[mk->mkchain[i][j]]);
      else if (mk->print_flag == PRINT_SER)
       fprintf(out,", %d ", mk->mkchain[i][j]);
      else if (mk->print_flag == PRINT_NAME)
        fprintf(out,", \"%s\" ", mk->state_name[mk->mkchain[i][j]]);
      }

    fprintf(out,"\n");   

   }

}


void create_mkchain(int n,MKCHAIN *mk)
{

  int ser,i;

  if (mk == (MKCHAIN *) NULL)
   {
    fprintf(stderr," *** ERROR: MKCHAIN NULL Pointer in create_mkchain()\n");
    exit(EXIT_FAILURE);
   }
  if (n < 1 || n > mk->nbr_chains)
   {
     fprintf(stderr," *** ERROR: n (%d) is less than 1 or greater than mk->nbr_chains (%d)\n",n,mk->nbr_chains);
     exit(EXIT_FAILURE);
   }

  /* get the initial point */

  ser = calc_state(mk->initial_probability,mk->nbr_states);
  mk->mkchain[n][0] = ser;

  /* create the chain */

  for (i=1 ; i <= mk->nbr_years ; i++)
   {
     ser = calc_state(mk->transition_probability[ser],mk->nbr_states);

     mk->mkchain[n][i] = ser;
   }

}


int calc_state(double *probability_table, int n)
 {
   int i;
   double random_val;
   double prob_partial_sum;

  if (probability_table == (double *) NULL || n < 1)
   {
     fprintf(stderr," *** ERROR: either NULL Pointer for probability_table or n < 1 (%d) in calc_state\n",n);
     exit(EXIT_FAILURE);
   }


  random_val = (double) rand() / (double) RAND_MAX;

  for (i = 1, prob_partial_sum = 0  ; i <= n ; i++)
   {
     prob_partial_sum += probability_table[i];

     if (random_val <= prob_partial_sum)
      return i;        

   }

   fprintf(stderr, " *** ERROR: Did not find new state -- random_val = %f  prob_partial_sum = %f  i = %d n = %d in calc_state()\n",
		random_val, prob_partial_sum, i, n);
   exit(EXIT_FAILURE);


  return 0;

 }

int calc_steady_state(double **a,double *b, int n, int max_iterations, double max_error, double *error,int debug)
{
  int nbr_iterations;
  int i,j,k;
  double **aa;
  
  int cmax_iteration; 
  double cmax_error, cerror;
  double row_average;
 


 if (a == (double **) NULL || b == (double *) NULL || n < 1)
  {
   fprintf(stderr," *** ERROR : NULL Pointer or n < 1 in calc_steady_state()\n");
   exit(EXIT_FAILURE);
  }

  aa = (double **) calloc((size_t)n+1,sizeof(double *));
  if (aa == (double **) NULL)
   {
    fprintf(stderr, " *** ERROR : NULL Pointer in calc_steady_state()\n");
    exit(EXIT_FAILURE);
   }

  for (i = 1 ; i <= n ; i++)
   {
     aa[i] = (double *) calloc((size_t)n+1,sizeof(double *));
     if (aa[i] == (double *) NULL)
      {
       fprintf(stderr, " *** ERROR : NULL Pointer in calc_steady_state()\n");
       exit(EXIT_FAILURE);
      }
   }

   if (max_iterations > 1)
    cmax_iteration = max_iterations;
   else
    cmax_iteration = DEFAULT_MAX_ITERATION;


   if (max_error > 0.0)
    cmax_error = max_error;
   else
    cmax_error = DEFAULT_MAX_ERROR;


   /* copy a into aa */

  for (j = 1 ; j <= n ; j++)
   for (k = 1 ; k <= n ; k++)
    aa[j][k] = a[j][k];


  /* iterate */

  for (i = 0 ; i < cmax_iteration ; i++)
   {
     multiply_array_by_array(aa,aa,aa,n,n,n);

     /* calculat the error */

     cerror = 0.0;

     /* rows should all have the same value in the columns*/

     for (j=1 ; j <= n ; j++)
      {
        row_average = 0;

        for (k=1 ; k <= n ; k++)
         row_average+=aa[k][j];

        row_average = row_average / (double) n;

        for (k=1 ; k <= n ; k++)
         cerror += (aa[k][j] - row_average)*(aa[k][j] - row_average);  /* sum of square of errors */

      }

    if (debug)
     {
      for (j=1 ; j <= n ; j++)
       {
        for (k=1 ; k <= n ; k++)
         fprintf(stderr,"%f ",aa[k][j]);
        fprintf(stderr,"\n");
       }
      fprintf(stderr,"i = %d  cerror = %f\n", i,cerror);


     }
    
    if (sqrt(cerror) < cmax_error)  /* see if done */
     break;

   }

  if (error != (double *) NULL) *error = sqrt(cerror);

  for (j=1 ; j < n+1 ; j++)
   {
    b[j] = aa[1][j];
   }

 if (debug){ fprintf(stderr,"$"); fflush(stderr);}

 for (j=1 ; j < n+1 ; j++)
  free((void *)aa[j]);

  free((void **)aa);
  
  return i;
}


void multiply_array_by_array(double **y, double **a, double **b,int nbr_r_a,int nbr_rc, int nbr_c_b)
{
  int i,j,k;
  double **val;

  if (y == (double **) NULL || a == (double **) NULL || b == (double **) NULL)
   {
    fprintf(stderr," *** ERROR: NULL Pointer in multiply_array_by_array()\n");
    exit(EXIT_FAILURE);
   }

  if (nbr_r_a < 1 || nbr_rc < 1 || nbr_c_b < 1)
   { 
    fprintf(stderr," *** ERROR: Illegal array dimension in mulitply_array_by_array()\n");
    exit(EXIT_FAILURE);
   }

  val = (double **)calloc((size_t) nbr_r_a+1, sizeof(double *));
  if (val == (double **) NULL)
   {
    fprintf(stderr," *** ERROR: NULL Pointer in multiply_array_by_array() a\n");
    exit(EXIT_FAILURE);
   }

  for (i = 0 ; i <= nbr_c_b ; i++)
   {
    val[i] = (double *)calloc((size_t) nbr_c_b+1, sizeof(double ));
    if (val[i] == (double *) NULL)
     {
      fprintf(stderr," *** ERROR: NULL Pointer in multiply_array_by_array() a\n");
      exit(EXIT_FAILURE);
     }
   }


  for (j = 1 ; j <= nbr_r_a ; j++)
   {
     for (k = 1 ; k <= nbr_c_b ; k++)
      {
        val[j][k] = 0.0;

        for (i = 1 ; i <= nbr_rc ; i++)
         val[j][k] += a[j][i] * b[i][k];

      }
   }

   
  for (j = 1 ; j <= nbr_r_a ; j++)
   {
     for (k = 1 ; k <= nbr_c_b ; k++)
      {
        y[j][k] = val[j][k];
      }

     free((void *)val[j]);
   }

  free((void *) val);
}





